#ifndef bqtStmtHH
#define bqtStmtHH

#include <list>

#include "autoptr"
#include "tokens.hh"
#include "expr.hh"

typedef autoptr<class stmt> stmtp;

class stmt_visitor
{
public:
    virtual ~stmt_visitor() {}
public:
    virtual void stmt_visit(stmtp e) = 0;
};
class stmt: public ptrable
{
    stmtp prev;
public:
    stmt();
    virtual ~stmt();
    
    void set_prev(stmtp p) { prev=p; }
    
    const stmtp& get_prev() const { return prev; }
    
    virtual bool MayErr() const { return true; }
    
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func) { func.stmt_visit(prev); }
    virtual void ForEachSubExpr(expr_visitor& ) {  }
};

class stmt_if: public stmt
{
public:
    exprp condition;
    stmtp thenp;
    std::list<
        std::pair<exprp, stmtp>
             > elseifp;
    stmtp elsep;
public:
    virtual void Compile();
    virtual bool MayErr() const
    {
        /* FIXME: Error may happen in elseifs too */
        return condition->MayErr();
    }

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};
class stmt_select: public stmt
{
public:
    exprp ref;
    std::list<
        std::pair<exprp, stmtp>
             > cases;
public:
    virtual void Compile();
    virtual bool MayErr() const { return ref->MayErr(); }
    
    bool IsEntirelyIntegerCase() const;

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};
class stmt_doloop: public stmt
{
public:
    exprp precondition;
    exprp postcondition;
    stmtp bodyp;
public:
    virtual void Compile();
    virtual bool MayErr() const { return (precondition&&precondition->MayErr())
                                     || (postcondition&&postcondition->MayErr()); }

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};
class stmt_for: public stmt
{
public:
    exprp var;
    exprp begin, end, step;
    stmtp bodyp;
public:
    virtual void Compile();
    virtual bool MayErr() const { return var->MayErr()
                                     || (begin&&begin->MayErr())
                                     || (end&&end->MayErr())
                                     || (step&&step->MayErr()); }

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};
class stmt_labelref: public stmt
{
public:
    QBtoken token;
    // GOTO, GOSUB or ON_ERROR, or RESTORE
    // Tokens:
    //  GOTO, GOSUB, ON_ERROR, RESTORE
    //  RETURN
    //  EXITDEF, EXITDO, EXITSUB, EXITFUNCTION, EXITFOR
    //  RESUME, RESUMENEXT
    //  STOP, END
    exprp label; // May only be expr_label
    // label only used by GOTO, GOSUB, ON_ERROR, RESTORE,
    //                    RESUME and RETURN
    
    stmt_labelref(QBtoken tok, exprp l=0) : token(tok), label(l) { }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};
class stmt_read: public stmt
{
    exprp target;
public:
    stmt_read(exprp t) : target(t) { }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};
class stmt_let: public stmt
{
    exprp target;
    exprp value;
public:
    stmt_let(exprp t, exprp v): target(t), value(v) { }
public:
    virtual void Compile();
    virtual bool MayErr() const { return target->MayErr() || value->MayErr(); }

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_print: public stmt // PRINT, LPRINT, PRINT #, WRITE #, PRINT USING.
{
public:
    exprp target;
    bool is_lprint; // Printer?
    exprp using_string;
    
    std::list<std::pair<QBtoken, exprp> > params;
    
    // Note: When parsing the USING format, SPC/TAB/CRLF do not consume a pattern.
public:
    stmt_print() : is_lprint(false) { }
    
    void SetFile(exprp p)       { target = p; }
    void SetLPRINT()            { is_lprint = true; }
    void SetFormat(exprp p)     { using_string = p; }
    void AppendExpr(exprp p)    { params.push_back(std::make_pair((QBtoken)0, p)); OptimizeTail(); }

    // Cursor position control:
    void AppendSPCfunc(exprp p) { params.push_back(std::make_pair(SPC, p)); }
    void AppendTABfunc(exprp p) { params.push_back(std::make_pair(TAB, p)); }
    void AppendTab()     { params.push_back(std::make_pair(TAB,  (expr*)0)); } // The ',' operator
    void AppendNewline() { params.push_back(std::make_pair(CRLF, (expr*)0)); }
    // For number printing:
    void AppendSpace()   { AppendExpr(new expr_string(" ")); }
    // For WRITE:
    void AppendQuoteCh() { AppendExpr(new expr_string("\"")); }
    void AppendCommaCh() { AppendExpr(new expr_string(",")); }
private:
    void OptimizeTail();
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_input: public stmt // INPUT, INPUT #, LINE INPUT, LINE INPUT #
{
public:
    exprp source;
    bool is_linput; // Line input? Line input has no prompt.
    exprp prompt;   // Default prompt is "?". Not used for LINEINPUT.
    bool keep_cursor;
    std::list<exprp> params;
public:
    stmt_input() : is_linput(false), keep_cursor(false) { }
    
    void SetFile(exprp p)   { source = p; }
    void SetLINEINPUT()     { is_linput = true; }
    void SetPrompt(exprp p) { prompt = p; }
    void SetKeepCursor()    { keep_cursor=true; }
    void AppendTarget(exprp p) { params.push_back(p); }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_fileopen: public stmt
{
    exprp filename;
    QBtoken mode;
    exprp mode_str;
    int access;
    int locking;
    exprp reclen;
public:
    stmt_fileopen(exprp fn, exprp mstr, int ac, int lk, exprp rl) 
        : filename(fn), mode_str(mstr), access(ac), locking(lk), reclen(rl)
    {
    }
    stmt_fileopen(exprp fn, QBtoken md, int ac, int lk, exprp rl) 
        : filename(fn), mode(md), access(ac), locking(lk), reclen(rl)
    {
    }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_redim: public stmt
{
    exprp target;
    std::vector<std::pair<exprp, exprp> > dimensions;
public:
    stmt_redim(exprp t, const std::vector<std::pair<exprp, exprp> >& d)
        : target(t), dimensions(d)
    {
    }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_erase: public stmt
{
    exprp target;
public:
    stmt_erase(exprp t) : target(t)
    {
    }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_subcall: public stmt
{
    std::string name;
    std::vector<exprp> params;
public:
    stmt_subcall(const std::string& n) : name(n) { }
    stmt_subcall(const std::string& n,
                 const std::vector<exprp>& p) : name(n), params(p) { }
public:
    const std::string& getname() const { return name; }

    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_label: public stmt
{
    long double label;
public:
    stmt_label() : label(0) { }
    stmt_label(long double f) : label(f) { }
    stmt_label(long long int i) : label(i) { }

public:
    virtual void Compile();
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_mid_s: public stmt
{
    exprp target;
    exprp start, length;
    exprp value;
public:
    stmt_mid_s(exprp t,exprp s,exprp l, exprp v): target(t),start(s),length(l),value(v) {}

public:
    virtual void Compile();
    virtual bool MayErr() const { return true; }

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};

class stmt_lib_stmt: public stmt
{
    QBtoken token;
    std::vector<exprp> params;
public:
    /* Note: Params must not be written into */
    
    stmt_lib_stmt(QBtoken t): token(t) {}
    stmt_lib_stmt(QBtoken tok, exprp a) : token(tok)
        { params.push_back(a); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b) : token(tok)
        { params.push_back(a); if(b)params.push_back(b); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b, exprp c) : token(tok)
        { params.push_back(a); if(b||c)params.push_back(b); if(c)params.push_back(c); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b, exprp c, exprp d) : token(tok)
        { params.push_back(a); if(b||c||d)params.push_back(b); if(c||d)params.push_back(c); if(d)params.push_back(d); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b, exprp c, exprp d, exprp e) : token(tok)
        { params.push_back(a); if(b||c||d||e)params.push_back(b); if(c||d||e)params.push_back(c);
          if(d||e)params.push_back(d); if(e)params.push_back(e); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b, exprp c, exprp d, exprp e, exprp f) : token(tok)
        { params.push_back(a); if(b||c||d||e||f)params.push_back(b); if(c||d||e||f)params.push_back(c);
          if(d||e||f)params.push_back(d); if(e||f)params.push_back(e); if(f)params.push_back(f); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b, exprp c, exprp d, exprp e, exprp f, exprp g) : token(tok)
        { params.push_back(a); if(b||c||d||e||f||g)params.push_back(b); if(c||d||e||f||g)params.push_back(c);
          if(d||e||f||g)params.push_back(d); if(e||f||g)params.push_back(e); if(f||g)params.push_back(f); if(g)params.push_back(g); }
    stmt_lib_stmt(QBtoken tok, exprp a, exprp b, exprp c, exprp d, exprp e, exprp f, exprp g, exprp h) : token(tok)
        { params.push_back(a); if(b||c||d||e||f||g||h)params.push_back(b); if(c||d||e||f||g||h)params.push_back(c);
          if(d||e||f||g||h)params.push_back(d); if(e||f||g||h)params.push_back(e); if(f||g||h)params.push_back(f);
          if(g||h)params.push_back(g); if(h)params.push_back(h); }
public:
    virtual void Compile();

    virtual void ForEachSubStmt(stmt_visitor& func);
    virtual void ForEachSubExpr(expr_visitor& func);
};



#endif
